/*! \file warpmac.h
\brief Header file for the WARPMAC Framework.

@version 12
@author Chris Hunter and Patrick Murphy

This header file contains the macros, function prototypes, and typedefs required for WARPMAC.
*/



/*! \mainpage WARP APIs
* @version Used with Reference Design v12
*
* \section change_sec CHANGELOG
* http://warp.rice.edu/trac/wiki/OFDMReferenceDesign/Changelog
*
* \section intro_sec Introduction
*
* This document is the collective API for WARPMAC (the MAC design framework),
* WARPPHY (the PHY driver/interface), RADIO CONTROLLER, CSMAMAC (the carrier-sensing random access
* reference), and the workshop MAC exercises.
*
* \section start_sec Getting Started
*
* Obtain the latest reference design: http://warp.rice.edu/trac/wiki/OFDMReferenceDesign .
* In addition to being in the repository, all the source code for WARPMAC is available in
* the reference design project and is configured to compile "as-is."
*
* \section hardcopy_sec Hardcopy
* You can download a hardcopy of this reference manual at:
* http://warp.rice.edu/WARP_API/refman.pdf
*
* \section licence_sec License
* http://warp.rice.edu/license/
*
*
*/



#ifndef WARPMAC_H
#define WARPMAC_H

#include "xstatus.h"
#include "warpphy.h"

//Define which radios get used below
//RADIOx_ADDR are defined by the radio controller driver
//Address of the first radio
#define FIRST_RADIO RADIO2_ADDR
//Address of the second radio
#define SECOND_RADIO RADIO3_ADDR


//Helper macros for constructing headers
#define htons(A) ((((Xuint16)(A) & 0xff00) >> 8) | (((Xuint16)(A) & 0x00ff) << 8))
#define NODEID_TO_ADDR(theID)	( theID & 0xFFFF )
#define ADDR_TO_NODEID(theAddr)	( theAddr & 0xFFFF )


#define MY_XEM_MAX_FRAME_SIZE 2000
#define EMACBUF_PING 0
#define EMACBUF_PONG 1

//Debug GPIO Pins
#define NONE	0
#define ONE		0x1
#define TWO		0x2
#define THREE	0x4
#define FOUR	0x8


//Masks for supported modulation schemes; equivalent to number of bits per symbol
#define BPSK	1
#define QPSK		2
#define QAM_16	4
#define QAM_64	6

//Constants used for packet headers specifying full-rate modulation per-packet
// This constant fills in an 8-bit field in the header
// Each 4-bit nibble corresponds to an antenna
#define HDR_FULLRATE_BPSK		(BPSK | (BPSK<<4))
#define HDR_FULLRATE_QPSK		(QPSK | (QPSK<<4))
#define HDR_FULLRATE_QAM_16	(QAM_16 | (QAM_16<<4))
#define HDR_FULLRATE_QAM_64	(QAM_64 | (QAM_64<<4))

//5 GHz Band
#define GHZ_5 0
//2.4 GHz Band
#define GHZ_2 1

//Borrowed from xemaclite.c; extracts the EtherType field from a received frame
#define XEmacLite_mGetReceiveDataLength(BaseAddress) ((XEmacLite_mReadReg((BaseAddress), XEL_HEADER_OFFSET + XEL_RXBUFF_OFFSET) >> XEL_HEADER_SHIFT) & (XEL_RPLR_LENGTH_MASK_HI | XEL_RPLR_LENGTH_MASK_LO))

//Shortcuts for xparameters.h constants
// Device IDs
#define EMACLITE_DEVICE_ID	 	XPAR_EMACLITE_0_DEVICE_ID
#define DMA_CTRL_DEVICE_ID		XPAR_DMACENTRAL_0_DEVICE_ID


#define DMADIR_IDLE			0
#define DMADIR_EMAC_TO_PHY	1
#define DMADIR_PHY_TO_EMAC	2

//The number of bytes in the header must be fixed and known by every node ahead of time
// It also must occupy an integral number of OFDM symbols (i.e. the base-rate symbols)
#define NUM_HEADER_BYTES 24

#define NUM_PAYLOAD_CRC_BYTES	4
//#define NUM_PAYLOAD_TAIL_BYTES	1 //Use for coded PHY
#define NUM_PAYLOAD_TAIL_BYTES	0	//Use for uncoded PHY

//Number of base rate OFDM symbols per packet
// Must correspond to base rate modulation, number of header bytes and header coding rate
//#define NUM_BASERATE_SYMBOLS	2 //Use for uncoded PHY
#define NUM_BASERATE_SYMBOLS	4 //Use for 1/2-rate coded header or uncoded BPSK header
#define NUM_TRAINING_SYMBOLS	2

///Structure contains the header contents of a packet
typedef struct {
	///Physical layer header struct
	phyHeader header;
	///Flag for protecting a Macframe. This is strictly optional and is only for use by user code; the framework ignores this field completely
	unsigned char remainingTx;
	///RIP isNew
} Macframe __attribute__((__aligned__(8)));

///Structure of miscellaneous control bits needed for correct operation of the MAC.
typedef struct {
	///Timeout time of system
	volatile unsigned int timeout;
	///Smallest time interval in system
	volatile unsigned int slotTime;
	///The node's self MAC address
	volatile unsigned char selfAddr;
	///Maximum number of retransmissions before dropping a packet
	volatile unsigned char maxReSend;
	///Maximum contention window index: [0, 2^(maxCW+4)-1]
	volatile unsigned char maxCW;
	///Current contention window index
	volatile unsigned char currentCW;
	///Whether the Emac is currently enabled
	volatile unsigned char enableEthernetInt;
	///Which of the PHY's buffers is currently associated with transmission
	volatile unsigned char txBuffIndex;
	///Which of the PHY's buffers is currently associated with reception
	volatile unsigned char rxBuffIndex;
	///Constellation order of the full rate symbols (#BPSK, #QPSK, #QAM_16, or #QAM_64)
	volatile unsigned char mod_fullRateA;
	//volatile unsigned char mod_fullRateB;
} Maccontrol;

///@brief Halts a currently running timer
///
///NOTE: This is a macro wrapper around ofdm_timer_stop()
#define warpmac_stopTimer() ofdm_timer_stop()

void nullCallback(void* param);
void warpmac_init();
void phyRx_goodHeader_handler();
void phyRx_badHeader_handler();
void warpmac_pollTimer();
int warpmac_clearTimer(unsigned char myTimer);
void warpmac_setTimerVal(Xuint32 clocks);
void warpmac_startTimer(unsigned char theTimer, unsigned char mode);
unsigned int randNum(unsigned int N);
unsigned int warpmac_carrierSense();
void warpmac_setTimer(int type);
int warpmac_startEmacXmit(Macframe* packet);
void warpmac_prepEmacForXmit(Macframe* packet);
void warpmac_setUpButtonCallback(void(*handler)());
void warpmac_setLeftButtonCallback(void(*handler)());
void warpmac_setRightButtonCallback(void(*handler)());
void warpmac_setMiddleButtonCallback(void(*handler)());
void warpmac_setTimerCallback(void(*handler)());
void warpmac_setEmacCallback(void(*handler)());
void warpmac_setGoodPacketCallback(void(*handler)());
void warpmac_setGoodHeaderCallback(void(*handler)());
void warpmac_setBadPacketCallback(void(*handler)());
void warpmac_setTxDoneCallback(void(*handler)());
void userIO_handler(void *InstancePtr);
void warpmac_incrementLEDLow();
void warpmac_incrementLEDHigh();
void warpmac_setBaseRate(unsigned char rate);
void warpmac_setFullRate(unsigned char rate);
void warpmac_prepPhyForXmit(Macframe* packet,unsigned char buffer);
void warpmac_startPhyXmit(unsigned char buffer);
void warpmac_setRxBuffer(Macframe* packet,unsigned int rxBuff);
void warpmac_setTxBuffer(unsigned int txBuff);
int warpmac_getRxBuffer();
int warpmac_getTxBuffer();
int warpmac_incrementResend(Macframe* packet);
int warpmac_addressedToMe(Macframe* packet);
int warpmac_addressedFromThem(Macframe* packet,unsigned char* addr);
int warpmac_addressedToThem(Macframe* packet,unsigned char* addr);
void warpmac_setMaxResend(unsigned int c);
void warpmac_setMaxCW(unsigned int c);
void warpmac_setMacAddr(unsigned char myID);
void warpmac_setTimeout(unsigned int time);
void warpmac_setSlotTime(unsigned int time);
int warpmac_inTimeout();
int warpmac_inBackoff();
int warpmac_getMyId();
void warpmac_enableCSMA();
void warpmac_disableCSMA();
unsigned char sevenSegmentMap(unsigned char x);
void warpmac_leftHex(unsigned char x);
void warpmac_rightHex(unsigned char x);
void warpmac_setDebugGPIO(unsigned char val);
void warpmac_enableEthernet();
void warpmac_disableEthernet();
void ErrorTrap(char *Message);
void warpmac_finishPhyXmit();
void warpmac_pollEthernet();
int warpmac_setAntennaMode(unsigned int txMode, unsigned int rxMode);
void warpmac_resetCurrentCW();
void warpmac_enableDummyPacketMode();
void warpmac_startPacketGeneration(unsigned int length, unsigned int interval);
void warpmac_stopPacketGeneration();
void warpmac_disableDummyPacketMode();

char warpmac_finishPhyRecv();

void warpmac_uartSendHandler(void *CallBackRef, unsigned int EventData);
//void warpmac_uartRecvHandler(void *CallBackRef, unsigned int EventData);
void warpmac_uartRecvHandler(void *CallBackRef);
void warpmac_setUartRecvCallback(void(*handler)());
int warpmac_sendRawEthernetPacket(void* packet, unsigned int numBytes);
void warpmac_pollPeripherals();

#ifdef WARP_FPGA_BOARD_V2_1
warpmac_middleHex(unsigned char x);	
unsigned char sevenSegmentMapHigh(unsigned char x);
unsigned char sevenSegmentMapLow(unsigned char x);
#endif

//FOR RTS/CTS

///timeout timer
#define TIMEOUT_TIMER 0
///backoff timer
#define BACKOFF_TIMER 1
///NAV timer //only used in RTS/CTS MAC
#define NAV 3
///User I/O polling timer
#define USERIO_TIMER 2
///User I/O polling timer
#define PKTGEN_TIMER 4

#define POLL_TIMER0
#define POLL_TIMER1
//#define POLL_TIMER2 //warpmac uses non-callback code for timer2, so skip the usual poll
#define POLL_TIMER3
//#define POLL_TIMER4 //warpmac uses non-callback code for timer4, so skip the usual poll

//#define POLL_TIMER5
//#define POLL_TIMER6
//#define POLL_TIMER7

#define REPLYTONEITHER 0x00
#define REPLYTOGOOD 0x01
#define REPLYTOBAD	0x02
#define REPLYTOBOTH 0x03

#define TIMERCLK_CYCLES_PER_USEC 80

//Poll the user I/O every 100 usec
#define USERIO_POLLRATE 100*TIMERCLK_CYCLES_PER_USEC


//JC Creating states for modulation rates (note different than MASKs!)
#define LOWEST_RATE  1
#define BPSK_MOD	   1
#define QPSK_MOD	   2
#define QAM16_MOD	   3
#define QAM64_MOD	   4
#define HIGHEST_RATE 3

//JC mrateMode states
#define BPSK_MODE	  		0
#define QPSK_MODE	  		1
#define QAM16_MODE  		2
#define BPSKRTS_MODE	  	3
#define QPSKRTS_MODE	  	4
#define QAM16RTS_MODE  	5
#define ARF_MODE			6
#define RRAA_MODE			7
#define RBAR_MODE			8
#define OAR_MODE			9

//JC rateDecision states
#define NOT_IN_USE  		0
#define CONS_WINDOW 		1
#define AVG_WINDOW  		2
#define SNR_BASED  		3

//JC rtsScheme states
#define DISABLED  		0
#define ENABLED	  		1
#define RRAA_TRIG  		2

//JC fairnessScheme
#define PACKET_BASED			0
#define TIME_SHARE_BASED  	1

//JC thresholds for average window mechanism
#define BPSK_PORI		4
#define QPSK_PORI		5
#define QPSK_PMTL		10
#define QAM16_PMTL	18

//JC speeds for channel training (change to coherenc time)
#define SP_STATIC		0
#define SP_09KMH		1
#define SP_25KMH		2
#define SP_50KMH		3
#define SP_100KMH		4

///Union for breaking unsigned shorts into 2 unsigned chars and vice versa
typedef union {
	unsigned short uint16;
	unsigned char uint8[2];
} UTime;

//BEGIN JC
unsigned char convertStatetoMask(unsigned char rate_state);
void increaseRate();
void decreaseRate();
//END JC

void initTX();
void sendCTS(Macframe* packet);
void sendDAT();
void sendBCD();
void dropPacket();
void updateTimer(unsigned char type, unsigned int newTime);

void unlockRXantenna();
void lockRXantenna();

//void setTXModMasks(unsigned char * newCSI);
void loadCSI(unsigned char * newCSI);

int rtsmac_phyRx_goodHeader_callback(Macframe* packet);
int rtsmac_phyRx_badHeader_callback();
int rtsmac_emacRx_callback(Xuint32 length, char* payload);
int rtsmac_timer_callback(unsigned char timerType);

void initFeedback();
void initDelays();
void initAddresses();
void initHeaders();
void rtsmac_left();
void rtsmac_right();
void rtsmac_middle();
void rtsmac_up();
void printState();
void clearPktCtr();
void printStatistics();

//JC
void prepHeader(Xuint32 length, char* payload);

#define ETHTYPE_CHANEST 0x9090

//Struct representing raw Ethernet header + packet contents
typedef struct {
	//Standard 14-byte Ethernet header (dest/src MAC addresses + EtherType) + 2-byte padding for 64-bit alignment
	unsigned char dstAddr[6];
	unsigned char srcAddr[6];
	unsigned short ethType;
	unsigned short zeroPadding;
} EthernetPktHeader;




#endif


